# Generated by Django 3.0.7 on 2021-03-15 23:22

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Room',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_when', models.DateTimeField(default=django.utils.timezone.now)),
                ('members', models.ManyToManyField(related_name='chat_rooms', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='RoomMembership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('archived', models.BooleanField(default=False)),
                ('ignore_notifications', models.BooleanField(default=False)),
                ('created_when', models.DateTimeField(default=django.utils.timezone.now)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='memberships', to='chit_chat.Room')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chat_room_memberships', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField()),
                ('created_when', models.DateTimeField(default=django.utils.timezone.now)),
                ('room', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='chit_chat.Room')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='chat_room_messages', to=settings.AUTH_USER_MODEL)),
                ('users_who_viewed', models.ManyToManyField(related_name='chat_room_messages_viewed', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



from rest_framework import viewsets, mixins
from django.db.models import Max

from chit_chat.models import Room
from chit_chat.serializers import RoomSerializer


class RoomViewSet(
    mixins.CreateModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    queryset = Room.objects.all()
    serializer_class = RoomSerializer

    def get_queryset(self):
        qs = super().get_queryset().filter(members=self.request.user)
        qs = qs.annotate(latest_message_time=Max('messages__created_when')).order_by('-latest_message_time')
        qs = qs.prefetch_related(
            'messages',
            'messages__users_who_viewed',
            'members',
        )
        return qs



import pytest
from django.contrib.sessions.models import Session
from django.contrib.auth import HASH_SESSION_KEY, SESSION_KEY, BACKEND_SESSION_KEY, get_user_model
from django.conf import settings
from django.test import override_settings
from channels.routing import get_default_application
from channels.testing import WebsocketCommunicator
from channels.db import database_sync_to_async
from asyncio.exceptions import TimeoutError

from chit_chat.consumers import ChatRoomConsumer
from testproject.testapp.factories import RoomFactory, UserFactory
from testproject.testapp.serializers import ChatTestSerializer

User = get_user_model()


@pytest.mark.django_db
@database_sync_to_async
def create_user(auth_backend=None):
    user = UserFactory()
    session_store = Session.get_session_store_class()()
    session_store[HASH_SESSION_KEY] = user.get_session_auth_hash()
    session_store[SESSION_KEY] = str(user.pk)
    backends = settings.AUTHENTICATION_BACKENDS
    if len(backends) > 1:  # pragma: no cover
        raise ValueError('Only one authentication backend should be set or else the auth_backend kwarg is required.')
    session_store[BACKEND_SESSION_KEY] = auth_backend or backends[0]
    session_store.create()
    return user, session_store.session_key


@pytest.mark.django_db
@database_sync_to_async
def create_room(**kwargs):
    return RoomFactory(**kwargs)


@pytest.mark.django_db
@database_sync_to_async
def get_messages_from_room(room):
    return list(room.messages.all().order_by('created_when').select_related('user'))


async def create_websocket_communicator(session_id):
    communicator = WebsocketCommunicator(
        get_default_application(),
        '/ws/chatroom/',
        headers=[('cookie'.encode(), f'sessionid={session_id}'.encode())]
    )
    connected, subprotocol = await communicator.connect(timeout=5)
    assert connected
    return communicator


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_connection_to_chat_room_requires_session_cookie():
    communicator = WebsocketCommunicator(get_default_application(), '/ws/chatroom/')
    connected, subprotocol = await communicator.connect(timeout=5)
    assert not connected

    user, session_key = await create_user()
    communicator = await create_websocket_communicator(session_key)
    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_user_is_able_to_connect_and_send_message_after_disconnecting():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    assert len(await get_messages_from_room(room)) == 0
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})

    resp = await communicator.receive_json_from()

    messages = await get_messages_from_room(room)
    assert len(messages) == 1
    assert messages[0].text == 'hello'
    assert messages[0].user == user

    assert resp['text'] == messages[0].text
    assert resp['time']

    await communicator.disconnect()

    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello 2'})

    resp = await communicator.receive_json_from(timeout=3)

    messages = await get_messages_from_room(room)
    assert len(messages) == 2
    assert messages[1].text == 'hello 2'
    assert messages[1].user == user

    assert resp['text'] == messages[1].text
    assert resp['time']

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_send_message_creates_new_message_object():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    assert len(await get_messages_from_room(room)) == 0
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})

    resp = await communicator.receive_json_from()

    messages = await get_messages_from_room(room)
    assert len(messages) == 1
    assert messages[0].text == 'hello'
    assert messages[0].user == user

    assert resp['text'] == messages[0].text
    assert resp['user'] == user.pk
    assert resp['time']

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_serializer_hook():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    assert len(await get_messages_from_room(room)) == 0
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})

    resp = await communicator.receive_json_from()
    resp.pop('time')
    assert resp == {'room': room.pk, 'text': 'hello', 'type': 'chat', 'user': user.pk}

    ckc_chat_settings = {
        'CHAT_MESSAGE_SERIALIZER': ChatTestSerializer,
    }
    with override_settings(CKC_CHAT_SERIALIZERS=ckc_chat_settings):
        await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})
        resp = await communicator.receive_json_from(timeout=3)
        assert resp == {'field_errors': {'text': [ChatTestSerializer.text_error]}}

    ckc_chat_settings['CHAT_MESSAGE_SERIALIZER'] = 'testapp.serializers.ChatTestSerializer'

    with override_settings(CKC_CHAT_SERIALIZERS=ckc_chat_settings):
        await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})
        resp = await communicator.receive_json_from()
        assert resp == {'field_errors': {'text': [ChatTestSerializer.text_error]}}

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_cannot_send_message_with_invalid_type():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'invalid', 'room': room.pk, 'text': 'hello'})

    resp = await communicator.receive_json_from()
    assert resp == {'field_errors': {'message_type': [f'Must be one of the following message types: {ChatRoomConsumer.MESSAGE_TYPES}']}}

    messages = await get_messages_from_room(room)
    assert len(messages) == 0

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_cannot_send_message_to_room_that_user_is_not_a_member_of():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    other_room = await create_room()
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_json_to({'message_type': 'chat', 'room': other_room.pk, 'text': 'hello'})

    resp = await communicator.receive_json_from(timeout=3)
    assert resp == {'non_field_errors': ['Cannot send messages to chat rooms that you are not a member of.']}

    messages = await get_messages_from_room(other_room)
    assert len(messages) == 0

    messages = await get_messages_from_room(room)
    assert len(messages) == 0

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_invalid_json_returns_useful_error_message():
    user, session_key = await create_user()
    room = await create_room(members=[user])
    communicator = await create_websocket_communicator(session_key)
    await communicator.send_to(bytes_data=b"hola\0")

    resp = await communicator.receive_json_from(timeout=5)
    assert resp == {'non_field_errors': ['Invalid JSON.']}

    messages = await get_messages_from_room(room)
    assert len(messages) == 0

    await communicator.disconnect()


@pytest.mark.asyncio
@pytest.mark.django_db
async def test_chat_message_proliferates_to_correct_chatroom():
    print(settings.CHANNEL_LAYERS)
    user, session_key = await create_user()
    other_user, other_session_key = await create_user()
    room = await create_room(members=[user, other_user])
    communicator = await create_websocket_communicator(session_key)
    other_communicator = await create_websocket_communicator(other_session_key)

    # Make sure that a user in a different room does not get the message
    unrelated_user, unrelated_session_key = await create_user()
    await create_room(members=[unrelated_user])
    unrelated_communicator = await create_websocket_communicator(unrelated_session_key)

    # User sends message
    await communicator.send_json_to({'message_type': 'chat', 'room': room.pk, 'text': 'hello'})

    # Other user receives message
    response = await other_communicator.receive_json_from()
    assert response['user'] == user.pk
    assert response['room'] == room.pk

    # User also receives message
    response = await communicator.receive_json_from()
    assert response['user'] == user.pk
    assert response['room'] == room.pk

    # This user should not receive a message
    try:
        await unrelated_communicator.receive_json_from()
    except TimeoutError:
        pass

    await communicator.disconnect()
    await other_communicator.disconnect()



from django.contrib.auth import get_user_model
from django.db.models import Count
from rest_framework import serializers, exceptions

from chit_chat.models import Room, Message
from .utils import PrimaryKeyWriteSerializerReadField


User = get_user_model()


class MessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = Message
        fields = (
            'id',
            'text',
            'created_when',
            'room',
            'user',
            'users_who_viewed',
        )


class ChatUserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = (
            'id',
            'first_name',
            'last_name',
            'avatar',
        )


class RoomSerializer(serializers.ModelSerializer):
    messages = MessageSerializer(read_only=True, many=True)
    members = PrimaryKeyWriteSerializerReadField(
        queryset=User.objects.all(),
        read_serializer=ChatUserSerializer,
        many=True,
    )

    class Meta:
        model = Room
        fields = (
            'id',
            'messages',
            'members',
        )

    def validate_members(self, users):
        requestor = self.context['request'].user
        non_requestor_users = [user for user in users if user != requestor]
        if len(non_requestor_users) < 1:
            raise exceptions.ValidationError('Must contain at least one user other than the requestor in this list.')
        non_requestor_users.append(requestor)
        return users

    def create(self, validated_data):
        member_pks = [member.pk for member in validated_data['members']]
        room = Room.objects.filter(members__in=member_pks).annotate(member_count=Count('members')).filter(member_count=len(member_pks)).first()
        return room or super().create(validated_data)



from rest_framework import serializers


#  TODO: This should be stored in the django-ckc package
class PrimaryKeyWriteSerializerReadField(serializers.PrimaryKeyRelatedField):
    def __init__(self, *args, **kwargs):
        read_serializer = kwargs.pop('read_serializer', None)
        assert read_serializer is not None, (
            'PrimaryKeyWriteSerializerReadField must provide `read_serializer` argument.'
        )
        super().__init__(*args, **kwargs)
        self.read_serializer = read_serializer

    # Related fields will not look up any of the object's values for normal primary
    # key serialization. This override forces the lookup of the entire object.
    def use_pk_only_optimization(self):
        return False

    def to_representation(self, value):
        return self.read_serializer(value, context=self.context).data



from channels.auth import AuthMiddlewareStack
from django.core.asgi import get_asgi_application
from channels.routing import ProtocolTypeRouter, URLRouter

import chit_chat.routing


application = ProtocolTypeRouter({
    'http': get_asgi_application(),
    'websocket': AuthMiddlewareStack(
        URLRouter(
            chit_chat.routing.websocket_urlpatterns
        )
    ),
})



from django.urls import reverse

from rest_framework.test import APITestCase

from testproject.testapp.factories import RoomFactory, MessageFactory, UserFactory


class TestChat(APITestCase):
    def setUp(self):
        self.user = UserFactory()
        self.client.force_authenticate(self.user)

    def test_room_create(self):
        room_data = {
            'members': [UserFactory().pk]
        }
        resp = self.client.post(reverse('room-list'), data=room_data)
        assert resp.status_code == 201

    def test_attempt_to_create_duplicate_rooms_results_in_original_being_returned(self):
        room_data = {
            'members': [UserFactory().pk]
        }
        resp = self.client.post(reverse('room-list'), data=room_data)
        assert resp.status_code == 201
        room_id = resp.json()['id']

        resp = self.client.post(reverse('room-list'), data=room_data)
        assert resp.status_code == 201
        assert room_id == resp.json()['id']

    def test_room_create_requires_another_user(self):
        room_data = {
            'members': [self.user.pk]
        }
        resp = self.client.post(reverse('room-list'), data=room_data)
        assert resp.status_code == 400
        assert resp.json() == {'members': ['Must contain at least one user other than the requestor in this list.']}

    def test_only_members_can_see_rooms(self):
        member_room = RoomFactory(members=[self.user])

        # Room that self.user is not a part of
        RoomFactory(members=[UserFactory()])

        resp = self.client.get(reverse('room-list'))
        assert resp.status_code == 200
        assert resp.json()['count'] == 1
        assert resp.json()['results'][0]['id'] == member_room.pk

    def test_get_room_messages(self):
        for _ in range(5):
            room = RoomFactory(members=[self.user])
            for _ in range(3):
                MessageFactory(room=room)

        resp = self.client.get(reverse('room-list'))
        assert resp.status_code == 200
        assert resp.json()['count'] == 5
        data = resp.json()['results']
        for room in data:
            assert len(room['messages']) == 3
            assert all(message['user'] for message in room['messages'])

    def test_rooms_are_ordered_by_latest_message(self):
        room_1 = RoomFactory(members=[self.user])
        MessageFactory(room=room_1)
        room_2 = RoomFactory(members=[self.user])
        MessageFactory(room=room_2)

        resp = self.client.get(reverse('room-list'))
        assert resp.status_code == 200
        assert resp.json()['count'] == 2
        data = resp.json()['results']
        assert data[0]['id'] == room_2.pk
        assert data[1]['id'] == room_1.pk

        MessageFactory(room=room_1)

        resp = self.client.get(reverse('room-list'))
        assert resp.status_code == 200
        assert resp.json()['count'] == 2
        data = resp.json()['results']
        assert data[0]['id'] == room_1.pk
        assert data[1]['id'] == room_2.pk
